@@ -1,13 +1,14 @@
 /* Prints out trees in human readable form.
    Copyright (C) 1992, 1993, 1994, 1995, 1996, 1998,
-   1999, 2000, 2001, 2002, 2003, 2004, 2005 Free Software Foundation, Inc.
+   1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007
+   Free Software Foundation, Inc.
    Hacked by Michael Tiemann (tiemann@cygnus.com)
 
 This file is part of GCC.
 
 GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
+the Free Software Foundation; either version 3, or (at your option)
 any later version.
 
 GCC is distributed in the hope that it will be useful,
@@ -16,9 +17,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GCC; see the file COPYING.  If not, write to
-the Free Software Foundation, 59 Temple Place - Suite 330,
-Boston, MA 02111-1307, USA.  */
+along with GCC; see the file COPYING3.  If not see
+<http://www.gnu.org/licenses/>.  */
 
 
 #include "config.h"
@@ -41,20 +41,21 @@ cxx_print_decl (FILE *file, tree node, int indent)
       return;
     }
 
-  if (!DECL_LANG_SPECIFIC (node))
+  if (!CODE_CONTAINS_STRUCT (TREE_CODE (node), TS_DECL_COMMON)
+      || !DECL_LANG_SPECIFIC (node))
     return;
   indent_to (file, indent + 3);
   if (TREE_CODE (node) == FUNCTION_DECL
       && DECL_PENDING_INLINE_INFO (node))
-    fprintf (file, " pending-inline-info " HOST_PTR_PRINTF,
+    fprintf (file, " pending-inline-info %p",
 	     (void *) DECL_PENDING_INLINE_INFO (node));
   if (TREE_CODE (node) == TYPE_DECL
       && DECL_SORTED_FIELDS (node))
-    fprintf (file, " sorted-fields " HOST_PTR_PRINTF,
+    fprintf (file, " sorted-fields %p",
 	     (void *) DECL_SORTED_FIELDS (node));
   if ((TREE_CODE (node) == FUNCTION_DECL || TREE_CODE (node) == VAR_DECL)
       && DECL_TEMPLATE_INFO (node))
-    fprintf (file, " template-info " HOST_PTR_PRINTF,
+    fprintf (file, " template-info %p",
 	     (void *) DECL_TEMPLATE_INFO (node));
 }
 
@@ -129,7 +130,7 @@ cxx_print_type (FILE *file, tree node, int indent)
 		 BINFO_N_BASE_BINFOS (TYPE_BINFO (node)));
       else
 	fprintf (file, " no-binfo");
-      
+
       fprintf (file, " use_template=%d", CLASSTYPE_USE_TEMPLATE (node));
       if (CLASSTYPE_INTERFACE_ONLY (node))
 	fprintf (file, " interface-only");
@@ -142,16 +143,22 @@ cxx_print_type (FILE *file, tree node, int indent)
 static void
 cxx_print_binding (FILE *stream, cxx_binding *binding, const char *prefix)
 {
-  fprintf (stream, "%s <" HOST_PTR_PRINTF ">",
+  fprintf (stream, "%s <%p>",
 	   prefix, (void *) binding);
 }
 
 void
 cxx_print_identifier (FILE *file, tree node, int indent)
 {
-  indent_to (file, indent);
+  if (indent == 0)
+    fprintf (file, " ");
+  else
+    indent_to (file, indent);
   cxx_print_binding (file, IDENTIFIER_NAMESPACE_BINDINGS (node), "bindings");
-  indent_to (file, indent);
+  if (indent == 0)
+    fprintf (file, " ");
+  else
+    indent_to (file, indent);
   cxx_print_binding (file, IDENTIFIER_BINDING (node), "local bindings");
   print_node (file, "label", IDENTIFIER_LABEL_VALUE (node), indent + 4);
   print_node (file, "template", IDENTIFIER_TEMPLATE (node), indent + 4);
@@ -165,7 +172,7 @@ cxx_print_xnode (FILE *file, tree node, int indent)
     case BASELINK:
       print_node (file, "functions", BASELINK_FUNCTIONS (node), indent + 4);
       print_node (file, "binfo", BASELINK_BINFO (node), indent + 4);
-      print_node (file, "access_binfo", BASELINK_ACCESS_BINFO (node), 
+      print_node (file, "access_binfo", BASELINK_ACCESS_BINFO (node),
 		  indent + 4);
       break;
     case OVERLOAD:
