@@ -1,6 +1,6 @@
 /* RunTime Type Identification
    Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-   2005
+   2005, 2006, 2007
    Free Software Foundation, Inc.
    Mostly written by Jason Merrill (jason@cygnus.com).
 
@@ -8,7 +8,7 @@ This file is part of GCC.
 
 GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
+the Free Software Foundation; either version 3, or (at your option)
 any later version.
 
 GCC is distributed in the hope that it will be useful,
@@ -17,9 +17,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GCC; see the file COPYING.  If not, write to
-the Free Software Foundation, 51 Franklin Street, Fifth Floor,
-Boston, MA 02110-1301, USA.  */
+along with GCC; see the file COPYING3.  If not see
+<http://www.gnu.org/licenses/>.  */
 
 #include "config.h"
 #include "system.h"
@@ -89,8 +88,8 @@ typedef enum tinfo_kind
   TK_POINTER_MEMBER_TYPE, /* abi::__pointer_to_member_type_info */
   TK_CLASS_TYPE,	/* abi::__class_type_info */
   TK_SI_CLASS_TYPE,	/* abi::__si_class_type_info */
-  TK_FIXED   		/* end of fixed descriptors. */
-  /* ... 		   abi::__vmi_type_info<I> */
+  TK_FIXED		/* end of fixed descriptors. */
+  /* ...		   abi::__vmi_type_info<I> */
 } tinfo_kind;
 
 /* A vector of all tinfo decls that haven't yet been emitted.  */
@@ -101,7 +100,6 @@ VEC(tree,gc) *unemitted_tinfo_decls;
    and are generated as needed. */
 static GTY (()) VEC(tinfo_s,gc) *tinfo_descs;
 
-static tree build_headof (tree);
 static tree ifnonnull (tree, tree);
 static tree tinfo_name (tree);
 static tree build_dynamic_cast_1 (tree, tree);
@@ -155,7 +153,7 @@ init_rtti_processing (void)
    virtual functions (TYPE_POLYMORPHIC_P), else just return the
    expression.  */
 
-static tree
+tree
 build_headof (tree exp)
 {
   tree type = TREE_TYPE (exp);
@@ -228,7 +226,7 @@ get_tinfo_decl_dynamic (tree exp)
   tree type;
   tree t;
 
-  if (exp == error_mark_node)
+  if (error_operand_p (exp))
     return error_mark_node;
 
   /* peel back references, so they match.  */
@@ -342,11 +340,10 @@ get_tinfo_decl (tree type)
   tree name;
   tree d;
 
-  if (COMPLETE_TYPE_P (type)
-      && TREE_CODE (TYPE_SIZE (type)) != INTEGER_CST)
+  if (variably_modified_type_p (type, /*fn=*/NULL_TREE))
     {
       error ("cannot create type information for type %qT because "
-	     "its size is variable",
+	     "it involves types of variable size",
 	     type);
       return error_mark_node;
     }
@@ -464,6 +461,9 @@ build_dynamic_cast_1 (tree type, tree expr)
   tree old_expr = expr;
   const char *errstr = NULL;
 
+  /* Save casted types in the function's used types hash table.  */
+  used_types_insert (type);
+
   /* T shall be a pointer or reference to a complete class type, or
      `pointer to cv void''.  */
   switch (tc)
@@ -616,6 +616,13 @@ build_dynamic_cast_1 (tree type, tree expr)
 		}
 	    }
 
+	  /* Use of dynamic_cast when -fno-rtti is prohibited.  */
+	  if (!flag_rtti)
+	    {
+	      error ("%<dynamic_cast%> not permitted with -fno-rtti");
+	      return error_mark_node;
+	    }
+
 	  target_type = TYPE_MAIN_VARIANT (TREE_TYPE (type));
 	  static_type = TYPE_MAIN_VARIANT (TREE_TYPE (exprtype));
 	  td2 = get_tinfo_decl (target_type);
@@ -674,9 +681,11 @@ build_dynamic_cast_1 (tree type, tree expr)
 	  if (tc == REFERENCE_TYPE)
 	    {
 	      tree bad = throw_bad_cast ();
+	      tree neq;
 
 	      result = save_expr (result);
-	      return build3 (COND_EXPR, type, result, result, bad);
+	      neq = c_common_truthvalue_conversion (result);
+	      return build3 (COND_EXPR, type, neq, result, bad);
 	    }
 
 	  /* Now back to the type we want from a void*.  */
@@ -813,13 +822,7 @@ tinfo_base_init (tinfo_s *ti, tree target)
     TREE_STATIC (name_decl) = 1;
     DECL_EXTERNAL (name_decl) = 0;
     DECL_TINFO_P (name_decl) = 1;
-    if (involves_incomplete_p (target))
-      {
-	TREE_PUBLIC (name_decl) = 0;
-	DECL_INTERFACE_KNOWN (name_decl) = 1;
-      }
-    else
-      set_linkage_according_to_type (target, name_decl);
+    set_linkage_according_to_type (target, name_decl);
     import_export_decl (name_decl);
     DECL_INITIAL (name_decl) = name_string;
     mark_used (name_decl);
@@ -853,7 +856,7 @@ tinfo_base_init (tinfo_s *ti, tree target)
 	 size_binop (MULT_EXPR,
 		     size_int (2 * TARGET_VTABLE_DATA_ENTRY_DISTANCE),
 		     TYPE_SIZE_UNIT (vtable_entry_type)));
-      
+
       ti->vtable = vtable_ptr;
     }
 
@@ -878,7 +881,7 @@ static tree
 generic_initializer (tinfo_s *ti, tree target)
 {
   tree init = tinfo_base_init (ti, target);
-  
+
   init = build_constructor_from_list (NULL_TREE, init);
   TREE_CONSTANT (init) = 1;
   TREE_INVARIANT (init) = 1;
@@ -937,7 +940,7 @@ ptm_initializer (tinfo_s *ti, tree target)
   init = tree_cons (NULL_TREE,
 		    get_tinfo_ptr (klass),
 		    init);
-  
+
   init = build_constructor_from_list (NULL_TREE, nreverse (init));
   TREE_CONSTANT (init) = 1;
   TREE_INVARIANT (init) = 1;
@@ -979,7 +982,6 @@ typeinfo_in_lib_p (tree type)
     {
     case INTEGER_TYPE:
     case BOOLEAN_TYPE:
-    case CHAR_TYPE:
     case REAL_TYPE:
     case VOID_TYPE:
       return true;
@@ -996,16 +998,16 @@ static tree
 get_pseudo_ti_init (tree type, unsigned tk_index)
 {
   tinfo_s *ti = VEC_index (tinfo_s, tinfo_descs, tk_index);
-  
+
   gcc_assert (at_eof);
   switch (tk_index)
     {
     case TK_POINTER_MEMBER_TYPE:
       return ptm_initializer (ti, type);
-      
+
     case TK_POINTER_TYPE:
       return ptr_initializer (ti, type);
-      
+
     case TK_BUILTIN_TYPE:
     case TK_ENUMERAL_TYPE:
     case TK_FUNCTION_TYPE:
@@ -1035,9 +1037,9 @@ get_pseudo_ti_init (tree type, unsigned tk_index)
 	VEC(tree,gc) *base_accesses = BINFO_BASE_ACCESSES (binfo);
 	tree base_inits = NULL_TREE;
 	int ix;
-          
+
 	gcc_assert (tk_index >= TK_FIXED);
-      
+
 	/* Generate the base information initializer.  */
 	for (ix = nbases; ix--;)
 	  {
@@ -1046,21 +1048,21 @@ get_pseudo_ti_init (tree type, unsigned tk_index)
 	    int flags = 0;
 	    tree tinfo;
 	    tree offset;
-	    
+
 	    if (VEC_index (tree, base_accesses, ix) == access_public_node)
 	      flags |= 2;
 	    tinfo = get_tinfo_ptr (BINFO_TYPE (base_binfo));
 	    if (BINFO_VIRTUAL_P (base_binfo))
 	      {
 		/* We store the vtable offset at which the virtual
-       		   base offset can be found.  */
+		   base offset can be found.  */
 		offset = BINFO_VPTR_FIELD (base_binfo);
 		offset = convert (sizetype, offset);
 		flags |= 1;
 	      }
 	    else
 	      offset = BINFO_OFFSET (base_binfo);
-	    
+
 	    /* Combine offset and flags into one field.  */
 	    offset = cp_build_binary_op (LSHIFT_EXPR, offset,
 					 build_int_cst (NULL_TREE, 8));
@@ -1117,7 +1119,7 @@ create_pseudo_type_info (int tk, const char *real_name, ...)
   va_start (ap, real_name);
 
   /* Generate the pseudo type name.  */
-  pseudo_name = alloca (strlen (real_name) + 30);
+  pseudo_name = (char *) alloca (strlen (real_name) + 30);
   strcpy (pseudo_name, real_name);
   strcat (pseudo_name, "_pseudo");
   if (tk >= TK_FIXED)
@@ -1145,6 +1147,10 @@ create_pseudo_type_info (int tk, const char *real_name, ...)
   ti->name = get_identifier (real_name);
   ti->vtable = NULL_TREE;
 
+  /* Pretend this is public so determine_visibility doesn't give vtables
+     internal linkage.  */
+  TREE_PUBLIC (TYPE_MAIN_DECL (ti->type)) = 1;
+
   va_end (ap);
 }
 
@@ -1156,29 +1162,29 @@ static unsigned
 get_pseudo_ti_index (tree type)
 {
   unsigned ix;
-  
+
   switch (TREE_CODE (type))
     {
     case OFFSET_TYPE:
       ix = TK_POINTER_MEMBER_TYPE;
       break;
-      
+
     case POINTER_TYPE:
       ix = TK_POINTER_TYPE;
       break;
-      
+
     case ENUMERAL_TYPE:
       ix = TK_ENUMERAL_TYPE;
       break;
-      
+
     case FUNCTION_TYPE:
       ix = TK_FUNCTION_TYPE;
       break;
-      
+
     case ARRAY_TYPE:
       ix = TK_ARRAY_TYPE;
       break;
-      
+
     case UNION_TYPE:
     case RECORD_TYPE:
       if (TYPE_PTRMEMFUNC_P (type))
@@ -1224,7 +1230,7 @@ get_pseudo_ti_index (tree type)
 		{
 		  /* too short, extend.  */
 		  unsigned len = VEC_length (tinfo_s, tinfo_descs);
-		  
+
 		  VEC_safe_grow (tinfo_s, gc, tinfo_descs, ix + 1);
 		  while (VEC_iterate (tinfo_s, tinfo_descs, len++, ti))
 		    ti->type = ti->vtable = ti->name = NULL_TREE;
@@ -1271,11 +1277,11 @@ static void
 create_tinfo_types (void)
 {
   tinfo_s *ti;
-  
+
   gcc_assert (!tinfo_descs);
 
   VEC_safe_grow (tinfo_s, gc, tinfo_descs, TK_FIXED);
-  
+
   push_nested_namespace (abi_node);
 
   /* Create the internal type_info structure. This is used as a base for
@@ -1306,15 +1312,15 @@ create_tinfo_types (void)
   create_pseudo_type_info (TK_ARRAY_TYPE, "__array_type_info", NULL);
   create_pseudo_type_info (TK_FUNCTION_TYPE, "__function_type_info", NULL);
   create_pseudo_type_info (TK_ENUMERAL_TYPE, "__enum_type_info", NULL);
-  
+
   /* Class type_info.  No additional fields.  */
   create_pseudo_type_info (TK_CLASS_TYPE, "__class_type_info", NULL);
-  
-  /* Single public non-virtual base class. Add pointer to base class. 
+
+  /* Single public non-virtual base class. Add pointer to base class.
      This is really a descendant of __class_type_info.  */
   create_pseudo_type_info (TK_SI_CLASS_TYPE, "__si_class_type_info",
-            build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
-            NULL);
+	    build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
+	    NULL);
 
   /* Base class internal helper. Pointer to base type, offset to base,
      flags.  */
@@ -1327,9 +1333,9 @@ create_tinfo_types (void)
     field = build_decl (FIELD_DECL, NULL_TREE, integer_types[itk_long]);
     TREE_CHAIN (field) = fields;
     fields = field;
-  
+
     ti = VEC_index (tinfo_s, tinfo_descs, TK_BASE_TYPE);
-    
+
     ti->type = make_aggr_type (RECORD_TYPE);
     ti->vtable = NULL_TREE;
     ti->name = NULL_TREE;
@@ -1351,10 +1357,10 @@ create_tinfo_types (void)
      This is really a descendant of __pbase_type_info.  */
   create_pseudo_type_info (TK_POINTER_MEMBER_TYPE,
        "__pointer_to_member_type_info",
-        build_decl (FIELD_DECL, NULL_TREE, integer_type_node),
-        build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
-        build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
-        NULL);
+	build_decl (FIELD_DECL, NULL_TREE, integer_type_node),
+	build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
+	build_decl (FIELD_DECL, NULL_TREE, type_info_ptr_type),
+	NULL);
 
   pop_nested_namespace (abi_node);
 }
@@ -1471,7 +1477,7 @@ emit_tinfo_decl (tree decl)
   if (DECL_NOT_REALLY_EXTERN (decl) && decl_needed_p (decl))
     {
       tree init;
-      
+
       DECL_EXTERNAL (decl) = 0;
       init = get_pseudo_ti_init (type, get_pseudo_ti_index (type));
       DECL_INITIAL (decl) = init;
