@@ -1,3 +1,6 @@
+// Derived from libstdc++/12048 by LJR <ljrittle@acm.org> with
+// reminder from Petur Runolfsson <peturr02@ru.is>.
+
 // Copyright (C) 2003 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
@@ -17,7 +20,6 @@
 // USA.
 
 #include <iostream>
-#include <cstdio>
 #include <testsuite_hooks.h>
 
 void
@@ -25,12 +27,13 @@ test01()
 {
   std::freopen("cin_unget-1.txt", "r", stdin);
 
-  char buf[2];
-  VERIFY( std::cin.rdbuf()->sgetn(buf, 2) == 2 );
-  int c1 = std::cin.rdbuf()->sungetc();
-  int c2 = std::fgetc(stdin);
-  VERIFY( c1 == std::char_traits<char>::to_int_type(buf[1]) );
-  VERIFY( c2 == c1 );
+  char c1;
+  char c2;
+  std::cin.get(c1);
+  std::cin.unget();
+  std::cin.get(c2);
+  VERIFY( std::cin.good() );
+  VERIFY( c1 == c2 );
 }
 
 int main(void)
