@@ -1,6 +1,6 @@
-// 2000-08-02 bkoz
+// 2003-04-30  Petur Runolfsson <peturr02@ru.is>
 
-// Copyright (C) 2000, 2001, 2002 Free Software Foundation, Inc.
+// Copyright (C) 2003 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
@@ -18,22 +18,30 @@
 // Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 // USA.
 
+#include <testsuite_hooks.h>
+#include <cstdio>
 #include <iostream>
 
-// libstdc++/2523
-void test03()
+void test02()
 {
   using namespace std;
-  ios_base::sync_with_stdio(false);
 
-  int i;
-  cin >> i;
-  cout << "i == " << i << endl;
+  bool test __attribute__((unused)) = true;
+
+  int c1 = fgetc(stdin);
+  int c2 = cin.rdbuf()->sputbackc(c1);
+  VERIFY( c2 == c1 );
+  
+  int c3 = fgetc(stdin);
+  VERIFY( c3 == c1 );
+  ungetc(c3, stdin);
+  
+  int c4 = cin.rdbuf()->sgetc();
+  VERIFY( c4 == c3 );
 }
 
-int 
-main()
+int main()
 {
-  test03();
+  test02();
   return 0;
 }
